from typing import (
    Any,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)

from django import forms

from reactivated import stubs

from .registry import JSON, Definitions, Thing, global_types, register

Override = TypeVar("Override")


class OptgroupMember(NamedTuple):
    name: str
    value: Union[str, int, bool, None]
    label: str
    selected: bool


Optgroup = Tuple[None, Tuple[OptgroupMember], int]


class BaseWidgetAttrs(NamedTuple):
    id: str
    disabled: stubs.Undefined[bool] = None
    required: stubs.Undefined[bool] = None
    placeholder: stubs.Undefined[str] = None


@register(forms.Widget)
class BaseWidget(NamedTuple):
    template_name: str

    name: str
    is_hidden: bool
    required: bool
    value: Optional[str]
    attrs: BaseWidgetAttrs

    @staticmethod
    def coerce_value(context: Any) -> Any:
        return context["value"]

    @classmethod
    def get_json_schema(
        Proxy: Type["BaseWidget"], instance: forms.Widget, definitions: Definitions,
    ) -> "Thing":
        from . import create_schema, named_tuple_schema

        # Subwidgets come to us as a class already.
        widget_class: Type[forms.Widget] = instance if isinstance(instance, type) else instance.__class__  # type: ignore[assignment]
        tag = f"{widget_class.__module__}.{widget_class.__qualname__}"

        if Proxy is BaseWidget:
            assert False, f"Unsupported widget {tag}"

        # TODO: this should really be done automatically for the instance we're wrapping.
        definition_name = tag  # f"{Proxy.__module__}.{Proxy.__qualname__}"

        if definition_name in definitions:
            return Thing(
                schema={"$ref": f"#/definitions/{definition_name}"},
                definitions=definitions,
            )

        base = named_tuple_schema(
            Proxy, definitions, definition_name=definition_name, exclude=["subwidgets"]
        )
        base = base.add_property("template_name", create_schema(Literal[widget_class.template_name], {}).schema)  # type: ignore[attr-defined]

        if subwidgets := get_type_hints(Proxy).get("subwidgets", None):
            if hasattr(subwidgets, "__annotations__"):
                subwidgets_tuple = Tuple[Any]
                subwidgets_tuple.__args__ = list(subwidgets.__annotations__.values())  # type: ignore[attr-defined]
                subwidget_keys = subwidgets.__annotations__.keys()
            else:
                subwidgets_tuple = subwidgets  # type: ignore[misc]
                subwidget_keys = [
                    str(index) for index in range(len(subwidgets.__args__))
                ]

            subwidgets_schema, definitions = create_schema(
                subwidgets_tuple, base.definitions
            )

            values_schema = {
                "type": "object",
                "required": [],
                "properties": {},
                "additionalProperties": False,
            }

            for subwidget_key, subwidget in zip(
                subwidget_keys, subwidgets_tuple.__args__  # type: ignore[attr-defined]
            ):
                value_schema = create_schema(subwidget, base.definitions).dereference()[
                    "properties"
                ]["value"]
                values_schema["properties"][subwidget_key] = value_schema  # type: ignore[index]
                values_schema["required"].append(subwidget_key)  # type: ignore[attr-defined]

            base = base._replace(definitions=definitions)
            base = base.add_property(
                "_reactivated_value_do_not_use", values_schema, optional=True
            )
            base = base.add_property("subwidgets", subwidgets_schema)

        base = base.add_property(
            "tag", create_schema(Literal[tag], base.definitions).schema
        )
        GlobalWidget: Dict[str, Any] = global_types.get("Widget", {"anyOf": [],})  # type: ignore[assignment]

        GlobalWidget = {
            **GlobalWidget,
            "anyOf": [*GlobalWidget["anyOf"], base.schema,],
        }
        global_types["Widget"] = GlobalWidget  # type: ignore[assignment]

        return base

    @classmethod
    def get_serialized_value(
        Proxy: Type["BaseWidget"], value: Any, schema: "Thing"
    ) -> JSON:
        from . import serialize

        widget_class = value.__class__
        context = value if isinstance(value, dict) else value._reactivated_get_context()

        # TODO: exclude value, and do it manually by calling get_value()
        # So call named_tuple serialized directly instead of serialize()
        serialized = serialize(context, schema, suppress_custom_serializer=True)
        serialized["tag"] = f"{widget_class.__module__}.{widget_class.__qualname__}"
        serialized["value"] = Proxy.coerce_value(context)

        if (subwidgets := get_type_hints(Proxy).get("subwidgets", None)) :
            subwidgets_to_enumerate = (
                subwidgets.__annotations__.values()
                if hasattr(subwidgets, "__annotations__")
                else subwidgets.__args__
            )

            for index, subwidget_class in enumerate(subwidgets_to_enumerate):
                serialized["subwidgets"][index][
                    "tag"
                ] = f"{subwidget_class.__module__}.{subwidget_class.__qualname__}"

        return serialized


class MaxLengthAttrs(BaseWidgetAttrs):
    maxlength: stubs.Undefined[str]


@register(forms.HiddenInput)
class HiddenInput(BaseWidget):
    template_name: Literal["django/forms/widgets/hidden.html"]
    type: Literal["hidden"]


@register(forms.TextInput)
class TextInput(BaseWidget):
    template_name: Literal["django/forms/widgets/text.html"]
    type: Literal["text"]
    attrs: MaxLengthAttrs


@register(forms.URLInput)
class URLInput(BaseWidget):
    template_name: Literal["django/forms/widgets/url.html"]
    type: Literal["url"]
    attrs: MaxLengthAttrs


class StepAttrs(BaseWidgetAttrs):
    step: stubs.Undefined[str]


@register(forms.NumberInput)
class NumberInput(BaseWidget):
    template_name: Literal["django/forms/widgets/number.html"]
    type: Literal["number"]
    attrs: StepAttrs


class TimeAttrs(BaseWidgetAttrs):
    format: stubs.Undefined[str]


@register(forms.TimeInput)
class TimeInput(BaseWidget):
    template_name: Literal["django/forms/widgets/time.html"]
    type: Literal["text"]
    attrs: TimeAttrs


class DateAttrs(BaseWidgetAttrs):
    format: stubs.Undefined[str]


@register(forms.DateInput)
class DateInput(BaseWidget):
    template_name: Literal["django/forms/widgets/date.html"]
    type: Literal["text"]
    attrs: DateAttrs


class DateTimeAttrs(BaseWidgetAttrs):
    format: str


@register(forms.DateTimeInput)
class DateTimeInput(BaseWidget):
    template_name: Literal["django/forms/widgets/datetime.html"]
    type: Literal["text"]
    attrs: DateTimeAttrs


class CheckAttrs(BaseWidgetAttrs):
    checked: stubs.Undefined[bool]


@register(forms.CheckboxInput)
class CheckboxInput(BaseWidget):
    template_name: Literal["django/forms/widgets/checkbox.html"]
    type: Literal["checkbox"]
    attrs: CheckAttrs
    value: bool  # type: ignore[assignment]

    @staticmethod
    def coerce_value(context: Any) -> bool:
        return context["attrs"].get("checked", False) is True


@register(forms.PasswordInput)
class PasswordInput(BaseWidget):
    template_name: Literal["django/forms/widgets/password.html"]
    type: Literal["password"]


@register(forms.EmailInput)
class EmailInput(BaseWidget):
    template_name: Literal["django/forms/widgets/email.html"]
    type: Literal["email"]


class TextareaAttrs(BaseWidgetAttrs):
    cols: str
    rows: str


@register(forms.Textarea)
class Textarea(BaseWidget):
    template_name: Literal["django/forms/widgets/textarea.html"]
    attrs: TextareaAttrs


@register(forms.Select)
class Select(BaseWidget):
    template_name: Literal["django/forms/widgets/select.html"]
    optgroups: List[Optgroup]

    @staticmethod
    def coerce_value(context: Any) -> Any:
        return context["value"][0]


class SelectMultipleAttrs(BaseWidgetAttrs):
    multiple: bool


@register(forms.SelectMultiple)
class SelectMultiple(BaseWidget):
    attrs: SelectMultipleAttrs
    optgroups: List[Optgroup]
    value: List[str]  # type: ignore[assignment]


@register(forms.ClearableFileInput)
class ClearableFileInput(BaseWidget):
    template_name: Literal["django/forms/widgets/clearable_file_input.html"]
    type: Literal["file"]
    checkbox_name: str
    checkbox_id: str
    is_initial: bool
    input_text: str
    initial_text: str
    clear_checkbox_label: str


class SelectDateWidgetValue(NamedTuple):
    year: Optional[Union[str, int]]
    month: Optional[Union[str, int]]
    day: Optional[Union[str, int]]


class SelectDateWidgetSubwidgets(NamedTuple):
    year: forms.Select
    month: forms.Select
    day: forms.Select


@register(forms.SelectDateWidget)
class SelectDateWidget(BaseWidget):
    template_name: Literal["django/forms/widgets/select_date.html"]
    subwidgets: SelectDateWidgetSubwidgets
    value: Any


@register(forms.SplitDateTimeWidget)
class SplitDateTimeWidget(BaseWidget):
    template_name: Literal["django/forms/widgets/splitdatetime.html"]
    subwidgets: Tuple[forms.DateInput, forms.TimeInput]
    value: Any



from django import forms

from reactivated.forms import Autocomplete, EnumChoiceField
from sample.server.apps.samples import models


class PlaygroundForm(forms.Form):
    choice_field = forms.ChoiceField(
        choices=(("", "- blank -"), ("first", "First"), ("second", "Second")),
        initial="first",
    )

    # TODO: figure out why adding this class makes the plugin work.
    # Something in
    # plugin.py#ctx.api.add_symbol_table_node(ctx.name, SymbolTableNode(GDEF, info))
    # causes a recursion error when using the plugin together with django-stubs
    # plugin.
    class Meta:
        pass


class ComposerForm(forms.ModelForm[models.Composer]):
    class Meta:
        model = models.Composer
        fields = "__all__"


ComposerFormSet = forms.modelformset_factory(
    model=models.Composer, form=ComposerForm, extra=0, min_num=1, validate_min=True
)


class OperaForm(forms.ModelForm[models.Opera]):
    class Meta:
        model = models.Opera
        fields = "__all__"
        widgets = {"composer": Autocomplete()}


OperaFormSet = forms.modelformset_factory(
    model=models.Opera, form=OperaForm, extra=0, min_num=1, validate_min=True
)


class StoryboardForm(forms.Form):
    char_field = forms.CharField()
    integer_field = forms.IntegerField()
    date_field = forms.DateField(widget=forms.SelectDateWidget)
    date_time_field = forms.DateTimeField(widget=forms.SplitDateTimeWidget,)
    choice_field = forms.ChoiceField(choices=((1, "One"), (2, "Two"),))
    enum_field = EnumChoiceField(enum=models.Opera.Style)
    boolean_field = forms.BooleanField()



import os
import subprocess
from typing import Any

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = "Generates all types and other client assets"

    def handle(self, *args: Any, **options: Any) -> None:
        entry_points = getattr(settings, "REACTIVATED_BUNDLES", ["index"])

        build_env = {
            **os.environ.copy(),
            "NODE_ENV": "production",
        }

        tsc_process = subprocess.Popen(
            ["./node_modules/.bin/tsc", "--noEmit"],
            stdout=subprocess.PIPE,
            env=build_env,
        )
        client_process = subprocess.Popen(
            ["node", "./node_modules/reactivated/build.client.js", *entry_points],
            stdout=subprocess.PIPE,
            env=build_env,
        )
        renderer_process = subprocess.Popen(
            ["node", "./node_modules/reactivated/build.renderer.js"],
            stdout=subprocess.PIPE,
            env=build_env,
        )

        tsc_output, tsc_error = tsc_process.communicate()
        client_output, client_error = client_process.communicate()
        renderer_ouput, renderer_error = renderer_process.communicate()

        if tsc_process.returncode != 0:
            raise CommandError("TypeScript error. Run 'tsc --noEmit' manually.")

        if client_process.returncode != 0  or renderer_process.returncode != 0:
            raise CommandError("esbuild errors")



from typing import Optional

import simplejson
from django.http import HttpRequest, JsonResponse
from django.urls import path

from .apps import get_schema
from .utils import get_attribute


def schema(request: HttpRequest, query: Optional[str] = None) -> JsonResponse:
    _schema = simplejson.loads(get_schema())

    if query:
        try:
            _schema = get_attribute(_schema, query.split("/"))
        except KeyError:
            _schema = {"error": "Invalid query"}

    return JsonResponse(_schema)


def client_bundle(request: HttpRequest, bundle: str) -> JsonResponse:
    import subprocess
    from django.views import static

    process = subprocess.Popen(
        ["node", "./node_modules/development/build.client.js"],
        stdout=subprocess.PIPE,
        # stdin=subprocess.PIPE,
    )
    process.wait()
    return static.serve(request, path=bundle, document_root="./dist/")


schema_views = [path("schema/", schema), path("schema/<path:query>/", schema)]

bundle_views = path("bundles/<str:bundle>", client_bundle)



import atexit
import logging
import re
import subprocess
import sys
import urllib.parse
from typing import Any, List, Optional

import simplejson
from django.conf import settings
from django.http import HttpRequest
from django.template.defaultfilters import escape

import requests_unixsocket

renderer_process: Optional[Any] = None
renderer_process_port: Optional[str] = None
logger = logging.getLogger("django.server")

# client_process =
# server_process =
# maybe_render_process =


def wait_and_get_port() -> Optional[str]:
    global renderer_process_port
    global renderer_process

    if renderer_process_port is not None:
        return renderer_process_port

    logger.info("Starting render process")
    entry_points = getattr(settings, "REACTIVATED_BUNDLES", ["index"])

    if renderer_process is None:
        renderer_process = subprocess.Popen(
            ["node", "./node_modules/.bin/renderer.js", *entry_points],
            encoding="utf-8",
            stdout=subprocess.PIPE,
        )

        def cleanup() -> None:
            # Pytest has issues with this, see https://github.com/pytest-dev/pytest/issues/5502
            # We can't use the env variable PYTEST_CURRENT_TEST because this happens
            # after running all tests and closing the session.
            # See: https://stackoverflow.com/questions/25188119/test-if-code-is-executed-from-within-a-py-test-session
            if "pytest" not in sys.modules:
                logger.info("Cleaning up renderer process")
            renderer_process.terminate()

        atexit.register(cleanup)

    output = ""

    for c in iter(lambda: renderer_process.stdout.read(1), b""):  # type: ignore[union-attr]
        output += c

        if match := re.match(r"RENDERER:([/.\w]+):LISTENING", output):
            renderer_process_port = match.group(1)
            return renderer_process_port
    assert False, "Could not bind to renderer"


def get_accept_list(request: HttpRequest) -> List[str]:
    """
    Given the incoming request, return a tokenized list of media
    type strings.

    From https://github.com/encode/django-rest-framework/blob/master/rest_framework/negotiation.py
    """
    header = request.META.get("HTTP_ACCEPT", "*/*")
    return [token.strip() for token in header.split(",")]


def should_respond_with_json(request: HttpRequest) -> bool:
    accepts = get_accept_list(request)

    return request.GET.get("format", None) == "json" or any(
        ["application/json" in content_type for content_type in accepts]
    )


def render_jsx_to_string(request: HttpRequest, context: Any, props: Any) -> str:
    respond_with_json = should_respond_with_json(request)

    payload = {"context": context, "props": props}
    data = simplejson.dumps(payload)
    headers = {"Content-Type": "application/json"}

    if "debug" in request.GET:
        return f"<html><body><h1>Debug response</h1><pre>{escape(data)}</pre></body></html>"
    elif (
        respond_with_json
        or "raw" in request.GET
        or getattr(settings, "REACTIVATED_SERVER", False) is None
    ):
        request._is_reactivated_response = True  # type: ignore[attr-defined]
        return data

    renderer_port = wait_and_get_port()

    session = requests_unixsocket.Session()
    socket = urllib.parse.quote_plus(renderer_port,)

    response = session.post(f"http+unix://{socket}", headers=headers, data=data)

    if response.status_code == 200:
        return response.text
    else:
        raise Exception(response.json()["stack"])
